// Copyright (c) 2003-2010 Christopher M. Kohlhoff
// Modifications (c) 2011 by Guilherme Maciel Ferreira / Intra2net AG
//
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)
#include "icmp/icmpv4header.h"

#include "boost_assert_handler.h"
#include "icmp/icmpdestinationunreachablemessage.h"
#include "icmp/icmpechoreplymessage.h"
#include "icmp/icmpechorequestmessage.h"
#include "icmp/icmpgenericmessage.h"

#include <logfunc.hpp>
using I2n::Logger::GlobalLogger;

using namespace std;
using boost::shared_ptr;

//-----------------------------------------------------------------------------
// Icmpv4Header
//-----------------------------------------------------------------------------

static const size_t HeaderSizeInBytes = 8;

Icmpv4Header::Icmpv4Header() :
    MessageFormat()
{
}

Icmpv4Header::Icmpv4Header(
        const Icmpv4Type type,
        const uint8_t code,
        const uint16_t checksum,
        const uint16_t identifier,
        const uint16_t sequence_number
) :
    MessageFormat()
{
    set_icmp_message_format( type );

    set_type( type );
    set_code( code );
    set_checksum( checksum );
    set_identifier( identifier );
    set_sequence_number( sequence_number );
}

Icmpv4Type Icmpv4Header::get_type() const
{
    return get_icmp_message_format()->get_type_v4();
}

void Icmpv4Header::set_type( const Icmpv4Type type )
{
    get_icmp_message_format()->set_type_v4( type );
} //lint !e1762

uint8_t Icmpv4Header::get_code() const
{
    return get_icmp_message_format()->get_code();
}

void Icmpv4Header::set_code( const uint8_t code )
{
    get_icmp_message_format()->set_code( code );
} //lint !e1762

uint16_t Icmpv4Header::get_checksum() const
{
    return get_icmp_message_format()->get_checksum();
}

void Icmpv4Header::set_checksum( const uint16_t checksum )
{
    get_icmp_message_format()->set_checksum( checksum );
} //lint !e1762

uint16_t Icmpv4Header::get_identifier() const
{
    return get_icmp_message_format()->get_identifier();
}

void Icmpv4Header::set_identifier( const uint16_t identifier )
{
    get_icmp_message_format()->set_identifier( identifier );
} //lint !e1762

uint16_t Icmpv4Header::get_sequence_number() const
{
    return get_icmp_message_format()->get_sequence_number();
}

void Icmpv4Header::set_sequence_number( const uint16_t sequence_number )
{
    get_icmp_message_format()->set_sequence_number( sequence_number );
} //lint !e1762

shared_ptr<IcmpMessage> Icmpv4Header::get_icmp_message_format() const
{
    BOOST_ASSERT( MessageFormat.get() != NULL );

    return MessageFormat;
}

void Icmpv4Header::set_icmp_message_format( const Icmpv4Type type )
{
    BOOST_ASSERT( MessageFormat.get() == NULL );

    if ( MessageFormat.get() == NULL )
    {
        switch ( type )
        {
            case Icmpv4Type_EchoReply:
                MessageFormat.reset( new IcmpEchoReplyMessage );
                break;
            case Icmpv4Type_EchoRequest:
                MessageFormat.reset( new IcmpEchoRequestMessage );
                break;
            case Icmpv4Type_DestinationUnreachable:
                MessageFormat.reset(
                        new IcmpDestinationUnreachableMessage
                );
                break;
            case Icmpv4Type_SourceQuench:
            case Icmpv4Type_Redirect:
            case Icmpv4Type_TimeExceeded:
            case Icmpv4Type_ParameterProblem:
            case Icmpv4Type_TimestampRequest:
            case Icmpv4Type_TimestampReply:
            case Icmpv4Type_InfoRequest:
            case Icmpv4Type_InfoReply:
            case Icmpv4Type_AddressRequest:
            case Icmpv4Type_AddressReply:
            case Icmpv4Type_Generic:
                MessageFormat.reset( new IcmpGenericMessage );
                break;
            case Icmpv4Type_InvalidLast:
            default:
                GlobalLogger.error() << "Invalid ICMPv4 message type " << static_cast<int>(type) << "!";
                BOOST_ASSERT( !"Try to set an invalid ICMPv4 message type" );
                break;
        }

        MessageFormat->init( type );
    }

    BOOST_ASSERT( MessageFormat.get() != NULL );
    BOOST_ASSERT( MessageFormat->get_type_v4() != Icmpv4Type_InvalidLast );
}

void Icmpv4Header::set_icmp_message_format( std::istream &is )
{
    // read the first byte, which contains the type of the ICMP message
    char first_byte;
    (void) is.read( &first_byte, 1 );

    // must keep the stream intact, so place the read byte back
    (void) is.putback( first_byte );

    // now select the message format for the given type
    Icmpv4Type header_type = static_cast<Icmpv4Type>( first_byte );
    set_icmp_message_format( header_type );
}

size_t Icmpv4Header::get_header_length() const
{
    return HeaderSizeInBytes;
}

std::istream& operator>>(
        std::istream &is,
        Icmpv4Header &header
)
{
    // select the message format which will parse the fields
    header.set_icmp_message_format( is );

    // delegate the parsing of the fields to a specialized decoder
    return header.get_icmp_message_format()->read( is );
}

std::ostream& operator<<(
        std::ostream &os,
        const Icmpv4Header &header
)
{
    // delegate the writing of the fields to a specialized encoder
    return header.get_icmp_message_format()->write( os );
}
